import errno
import logging
import multiprocessing as mp
import queue
import signal
import traceback
from typing import Any, Callable, Iterable, List

from ._joboutput import JobOutput

logger = logging.getLogger(__name__)


class MpRunner(object):
    """
    A general purpose multiprocessing runner class. It can be used
    for purely independent jobs without shared writable objects.

    Note: This class is not thread safe when multiple workers are writing the same objects.
    """

    def __init__(
        self,
        num_workers: int,
    ):
        self.num_workers = self._parse_num_workers(num_workers)

    @staticmethod
    def _parse_num_workers(num_workers: int) -> int:
        """parse the number of workers

        num_workers :param: number of workers intended to use
        :return: parsed number of workers to use,
        """
        if 0 <= num_workers <= 1:
            return 0
        elif num_workers < 0:
            return mp.cpu_count() - 1
        else:
            return int(num_workers)

    def _parallel_exec(self, jobs: Iterable, exec_func: Callable) -> List[JobOutput]:
        """handle multiple defined jobs via multiprocessing

        jobs :param: An iterable list that defines the jobs
        :return: None
        """
        logger.info(f"Start running the jobs on {self.num_workers} CPUs.")

        manager = mp.Manager()
        job_queue = manager.Queue()
        for i, job in enumerate(jobs):
            job_queue.put((i, job))
        result_queue = manager.Queue()

        workers = []
        for _ in range(self.num_workers):
            proc = mp.Process(
                target=self._executor_hook, args=(job_queue, result_queue, exec_func)
            )
            proc.start()
            workers.append(proc)

        # synchronize the forked workers and handle keyboard interrupt in main process
        try:
            for worker in workers:
                worker.join()
        except KeyboardInterrupt:
            for worker in workers:
                worker.terminate()
                worker.join()

        _results = []
        while not result_queue.empty():
            _results.append(result_queue.get(block=False))

        return _results

    @staticmethod
    def executor(jobid: int, job: Any, exec_func: Callable) -> JobOutput:
        """job handler"""
        # noinspection PyBroadException
        try:
            exec_rc = exec_func(job)
            return JobOutput(
                jobid=jobid, run_ok=True, result=dict(output=exec_rc, msg="")
            )
        except Exception:
            msg = traceback.format_exc()
            return JobOutput(
                jobid=jobid, run_ok=False, result=dict(output=None, msg=msg)
            )

    def run(self, jobs: Iterable, exec_func: Callable) -> List[JobOutput]:
        """the main func to run the jobs"""
        # global multiprocessing_runner
        # multiprocessing_runner = self
        results = None

        if self.num_workers > 1:
            try:
                results = self._parallel_exec(jobs, exec_func)
            except IOError as ie:
                if ie.errno == errno.EPIPE:
                    raise Exception("Broken pipe error, interrupted.")
        else:
            # sequentially run the job when num_workers <= 1
            results = [self.executor(i, job, exec_func) for i, job in enumerate(jobs)]

        return results

    def _executor_hook(
        self, job_queue: mp.Queue, result_queue: mp.Queue, exec_func: Callable
    ):
        # ignore the CTRL+C signaling
        signal.signal(signal.SIGINT, signal.SIG_IGN)

        while not job_queue.empty():
            # noinspection PyBroadException
            try:
                jobid, job = job_queue.get(block=False)
                result_queue.put(self.executor(jobid, job, exec_func))
            except queue.Empty:
                pass
            except Exception:
                traceback.print_exc()



from typing import Any, Dict, Literal, Optional, Union

import numpy as np
import pandas as pd
from anndata import AnnData
from numpy.random import RandomState
from scanpy.tools._utils import _choose_representation
from scipy.sparse import issparse
from sklearn.metrics import pairwise_distances
from sklearn.utils import check_array, check_random_state

from ._utils import random_feats

_KernelCapable = Literal[
    "additive_chi2",
    "chi2",
    "linear",
    "poly",
    "polynomial",
    "rbf",
    "laplacian",
    "sigmoid",
    "cosine",
]

SHIFT_INVARIANT_KERNEL = ["rbf"]


def kernel_mean_embedding(
    adata: AnnData,
    partition_key: str,
    X_anchor: np.ndarray,
    n_pcs: Optional[str] = None,
    use_rep: Optional[str] = None,
    random_state: Union[None, int, RandomState] = 0,
    method: Literal["exact", "approx"] = "exact",
    kernel: _KernelCapable = "rbf",
    kernel_kwds: Dict[str, Any] = {},
    n_jobs: Optional[int] = None,
    key_added: Optional[str] = None,
    copy=False,
) -> Optional[AnnData]:
    """\
    Landmark-based Kernel Mean Embedding

    *Kernel mean embedding* (KME) is a method used in machine learning and statistics
    to map probability distribution into a Reproducing Kernel Hilbert Space (RKHS).
    After embedding distributions into a RKHS, we can work with distributions in a more
    tractable way.

    Here, we applied the `KME` to sample distributions. By evaluating the transformed
    distributions in the RKHS at the landmark cells, we align them at the same landmark
    space, and generate a sample-by-landmark kernel score matrix as sample embeddings.

    .. note::
       To know more about the background of kernel mean embedding, please refer to this book :cite:`kme2017`.

    Parameters
    ----------
    adata
        Annotated data of type `anndata.AnnData`
    partition_key
        Column key in the field of `adata.obs` indicating the sample ids
    X_anchor
        An numpy ndarray as embeddings of landmark cells. **Note**: `X_anchor`
        should have the same dimension with the representation to be used in the `adata`.
    n_pcs
        Use this many PCs. If `n_pcs==0` use `.X` if `use_rep is None`
    use_rep
        Use the indicated representation. `'X'` or any key for `.obsm` is valid.
        If `None`, the representation is chosen automatically:
        For `.n_vars` < 50, `.X` is used, otherwise 'X_pca' is used.
        If 'X_pca' is not present, it's computed with default parameters.
    random_state
        A random seed which supports an `int`, `None` and `numpy.random.RandomState`
    method
        method to use for kernel calculation, supporting two options `exact` an `approx`.
        `exact` means directly calculate the kernel matrix. `approx` means that
        the kernel is approximated by random fourier features :cite:`rahimi2007random`.
    kernel
        kernels to use. We support many kernels like `rbf`, `linear`, and `laplacian`.
        For more kernels, see the global var: `sclkme.tl._kme._KernelCapable`.
    kernel_kwds
        A dic holding extra kwargs for the function of `sklearn.metrics.pairwise_kernels`
        in the calculation of kernels between cells.
    n_jobs
        number of cpus to use in the calculation of kernels. `n_jobs = 0 or None` means
        only use one CPU. `n_jobs > 0 and n_jobs < MAX_CPUS_AVAILABLE` will use that
        number of cpus. `n_jobs = -1` means using all CPUs. For more details, check the
        document of `sklearn.metrics.pairwise_kernels`.
    key_added
        If not specified, the results including params are stored in `.uns['kme']` as
        a dict, and sample embeddings are stored in `.obs['X_kme']`.
        If specified, the results are stored in `.uns[f{key_added}_kme]`, and the
        sample embeddings are stored in `.obsm['X_kme']`.
    copy
        Return a copy instead of writing the results into `adata`.

    Returns
    -------
    Depending on `copy`, updates or returns `adata` with the following fields.

    **X_kme** : `adata.obsm` field
        sample embeddings of data
    **kme** or **{key_added}_kme**: `adata.uns` field
        results from kernel mean embeddings, including both parameters and sample embeddings

    Example
    -------
    >>> import sclkme
    >>> import scanpy as sc

    Load annotated dataset:

    >>> adata = sc.datasets.pbmc3k_processed()

    Run kernel mean embedding to generate sample embeddings

    .. note::
       Here, we just use the `louvain` field in `adata.obs` as **sample ids** for illustration.
       In your application, you need to change it to your real sample ids.

    >>> sclkme.tl.sketch(adata, n_sketch=128, use_rep="X_pca")
    >>> X_anchor = adata[adata.obs['sketch']].obsm['X_pca'].copy()
    >>> sclkme.tl.kernel_mean_embedding(adata, partition_key='louvain', use_rep="X_pca", X_anchor=X_anchor)
    """
    # first we run the sketching algorithm to find the anchors
    adata = adata.copy() if copy else adata

    if key_added is None:
        key_added = "kme"
    else:
        key_added = key_added + "_kme"

    adata.uns[key_added] = {}

    kme_dict = adata.uns[key_added]

    kme_dict["partition_key"] = partition_key
    kme_dict["params"] = {"kernel": kernel, "method": method}
    kme_dict["params"]["random_state"] = random_state
    if kernel_kwds:
        kme_dict["params"]["kernel_kwds"] = kernel_kwds

    X = _choose_representation(adata, use_rep=use_rep, n_pcs=n_pcs, silent=True)
    if issparse(X):
        X = X.toarray()
    if use_rep is not None:
        kme_dict["params"]["use_rep"] = use_rep
    if n_pcs is not None:
        kme_dict["params"]["n_pcs"] = n_pcs

    # check the input anchor matrix
    X_anchor = check_array(
        X_anchor, dtype=np.float32, ensure_2d=True, accept_sparse=False
    )
    if X_anchor.shape[1] != X.shape[1]:
        raise ValueError(
            f"the shape: {X_anchor.shape} of input anchor matrix \
            cannot match the data representation shape in `adata`."
        )

    if method == "exact":
        from sklearn.metrics import pairwise_kernels

        # compute the kernels between X and X_anchor
        if kernel == "rbf":
            X_kme = rbf_kernel(
                X, X_anchor, gamma=kernel_kwds.get("gamma", 1), n_jobs=n_jobs
            )
        elif kernel == "laplacian":
            X_kme = laplacian_kernel(
                X=X, Y=X_anchor, gamma=kernel_kwds.get("gamma", 1), n_jobs=n_jobs
            )
        else:
            X_kme = pairwise_kernels(
                X=X, Y=X_anchor, metric=kernel, n_jobs=n_jobs, **kernel_kwds
            )

        # make the kernel matrix sparse if necessary
        mask = X_kme > 1e-14
        X_kme[~mask] = 0

    elif method == "approx":
        if kernel not in SHIFT_INVARIANT_KERNEL:
            raise ValueError(
                f"Only shift invariant kernels: [{','.join(SHIFT_INVARIANT_KERNEL)}] "
                "support `method=approx` for kernel mean embedding."
            )
        kme_dict["params"]["kernel_kwds"]["gamma"] = kernel_kwds.get("gamma", 1)
        kme_dict["params"]["kernel_kwds"]["D"] = kernel_kwds.get("D", 2000)

        random_state = check_random_state(random_state)
        phi_X = random_feats(
            X,
            gamma=kme_dict["params"]["kernel_kwds"]["gamma"],
            D=kme_dict["params"]["kernel_kwds"]["D"],
            random_state=random_state,
        )
        phi_X_anchor = random_feats(
            X_anchor,
            gamma=kme_dict["params"]["kernel_kwds"]["gamma"],
            D=kme_dict["params"]["kernel_kwds"]["D"],
            random_state=random_state,
        )
        # outer multiplication
        X_kme = phi_X @ phi_X_anchor.T

    adata.obsm["X_kme"] = X_kme  # save the results

    # save the resulting embeddings based on partition_key
    df_partition = pd.DataFrame(
        adata.obsm["X_kme"].copy(),
        columns=[f"X_kme_{i}" for i in range(adata.obsm["X_kme"].shape[1])],
    )
    df_partition[partition_key] = adata.obs[partition_key].tolist()
    df_partition = df_partition.groupby(partition_key, sort=False).aggregate(np.mean)
    kme_dict[f"{partition_key}_kme"] = df_partition

    return adata if copy else None


def rbf_kernel(X, Y=None, gamma=None, normalized=True, n_jobs=None):
    """Compute the rbf (gaussian) kernel between X and Y

        K(X, Y) = exp(-gamma ||x - y||^2 / Den)

    for each pair of rows in X and y in Y.
    Note: Den := np.max(Dsq, axis=0) / 4

    """
    Dsq = pairwise_distances(X, Y, metric="euclidean", squared=True, n_jobs=n_jobs)
    Den = np.max(Dsq, axis=0) / 4 if normalized else 1

    if gamma is None:
        gamma = 1.0 / X.shape[1]

    K = np.exp(-gamma * Dsq / Den)
    return K


def laplacian_kernel(X, Y, gamma=None, normalized=True, n_jobs=None):
    """Compute the laplacian kernel between X and Y.

        K(X, Y) = exp(-gamma ||x - y||_1 / Den)

    for each pair of rows x in X and y in Y.
    """
    D_manhattan = pairwise_distances(X, Y, metric="manhattan", n_jobs=n_jobs)
    Den = np.max(D_manhattan, axis=0) / 4 if normalized else 1

    if gamma is None:
        gamma = 1.0 / X.shape[1]

    K = np.exp(-gamma * D_manhattan / Den)
    return K



import functools
import gc
import os
from typing import Any, Callable, Dict, List, Optional, Union

import anndata
import pandas as pd
import readfcs
from anndata import AnnData
from geosketch import gs
from numpy import random
from PyCytoData import PyCytoData
from rich.console import Console
from scanpy._utils import sanitize_anndata

from sclkme.utils import MpRunner

logger = Console(log_path=False)


def prepare_cytometry_data(
    data_dir: str,
    metadata: str,
    filename_col: str,
    lineage_channels: List,
    groupby: Optional[str] = None,
    down_sample: Optional[int] = None,
    num_workers: int = -1,
    random_state: Union[None, int, random.RandomState] = 0,
    gate_kwds: Dict[str, Any] = {},
    verbose: bool = False,
) -> Union[AnnData, Dict[str, AnnData]]:
    # read the metadata first
    meta_df = pd.read_csv(metadata, header=0)

    # check if key exists
    assert (filename_col in meta_df.columns) and (
        groupby in meta_df.columns if groupby is not None else True
    ), f"the meta data file: {metadata} doesn't have the filename column: {filename_col}"

    task_handler = functools.partial(
        read_data,
        data_dir=data_dir,
        filename_col=filename_col,
        lineage_channels=lineage_channels,
        down_sample=down_sample,
        replace=False,
        random_state=random_state,
        gate_kwds=gate_kwds,
        verbose=verbose,
    )
    logger.log(
        f"Reading and preprocessing {len(meta_df[filename_col])} files. It may take a few minutes..."
    )

    adata = None
    if groupby is not None:
        adata = dict()
        for grp_name, grp in meta_df.groupby(groupby, sort=False):
            adata[grp_name] = _read_handler(
                grp, func=task_handler, num_workers=num_workers
            )
    else:
        adata = _read_handler(meta_df, func=task_handler, num_workers=num_workers)
    logger.log("Your data is converted into `AnnData`.")

    if adata is None:
        raise RuntimeError(
            "Reading data failed, please check the input and try again..."
        )

    logger.log("Done!")

    return adata


# read the data
def _read_handler(rows_df: pd.DataFrame, func: Callable, num_workers: int):
    # start multiprocessing pool to handle jobs
    multiprocessing_runner = MpRunner(num_workers)
    results = multiprocessing_runner.run(
        [rows_df.iloc[i, :] for i in range(len(rows_df.index))], func
    )

    assert len(rows_df.index) == len(results), "input job length != result length"

    adata_list = [None for _ in range(len(rows_df.index))]
    for res in results:
        if res.run_ok:
            if res.result["output"] is None:
                raise RuntimeError("resulting adata is None, exit...")
            adata_list[res.jobid] = res.result["output"]
        else:
            raise RuntimeError(
                f"Error occurred in reading the processing {res.result['msg']}"
            )

    for jobid, adata in enumerate(adata_list):
        if adata is None:
            raise RuntimeError(f"job: {jobid} is not correctly handled.")

    # merge the data
    adata = anndata.concat(adata_list, merge="same", index_unique="-", axis=0)  # type: ignore
    adata.obs.reset_index(drop=True, inplace=True)
    sanitize_anndata(adata)

    del adata_list, results
    gc.collect()

    return adata


def read_data(
    row: pd.Series,
    data_dir: str,
    filename_col: str,
    lineage_channels: str,
    down_sample: Optional[int] = None,
    replace: bool = False,
    random_state: Union[None, int, random.RandomState] = 0,
    gate_kwds: Dict[str, Any] = None,
    verbose: bool = False,
):
    if verbose:
        logger.log(f"Reading and preprocessing file: {row[filename_col]}")

    if row[filename_col].endswith(".fcs"):
        adata = readfcs.read(os.path.join(data_dir, row[filename_col]), reindex=False)
        assert (
            "channel" in adata.var.columns
        ), f"the col: channel cannot be found in the fcs file: {row[filename_col]}"
        adata.var.set_index("channel", inplace=True)
    elif row[filename_col].endswith(".csv"):
        adata = anndata.read_csv(os.path.join(data_dir, row[filename_col]))
    elif row[filename_col].endswith(".tsv"):
        adata = anndata.read_text(
            os.path.join(data_dir, row[filename_col]), delimiter="\t"
        )
    else:
        raise ValueError(
            "the file extension is not understood for reading, please use .fcs, .csv and .tsv file."
        )

    channels = adata.var.index.tolist()
    expr = PyCytoData(
        expression_matrix=adata.X.copy(),
        channels=channels,
        lineage_channels=lineage_channels,
    )
    if gate_kwds is not None:
        expr.preprocess(**gate_kwds, verbose=verbose)
    adata = anndata.AnnData(X=expr.expression_matrix, obs=row.to_dict(), var=adata.var)

    if down_sample is not None and adata.n_obs >= down_sample:
        if verbose:
            logger.log(f"downsample to {down_sample} cells.")
        gs_idx = gs(
            adata[:, lineage_channels].X,
            down_sample,
            replace=replace,
            seed=random_state,
        )
        adata = adata[gs_idx].copy()

    return adata



from typing import Union

import numpy as np
from numpy.random import RandomState


def random_feats(
    X: np.ndarray,
    gamma: Union[int, float] = 1,
    D: int = 2000,
    random_state: Union[None, int, RandomState] = None,
):
    """
    Computes random fourier frequency features: https://papers.nips.cc/paper/2007/hash/013a006f03dbc5392effeb8f18fda755-Abstract.html

    Parameters
    X: np.ndarray
        array of input data (dimensions = cells x features)
    gamma: Union([int, float]) (default = 1)
        scale for standard deviation of the normal distribution
    D: int (default = 2000):
        dimensionality of the random Fourier frequency features, D/2 sin and D/2 cos basis
    frequency_seed: int (default = None):
        random state parameter
    ----------
    Returns
    phi: np.ndarray
        random Fourier frequency features (dimensions = cells x D)
    ----------
    """
    scale = 1 / gamma
    d = int(np.floor(D / 2))

    if isinstance(random_state, RandomState):
        W = random_state.normal(scale=scale, size=(X.shape[1], d))
    else:
        # set global random seed if the input is an integer
        np.random.seed(random_state)
        W = np.random.normal(scale=scale, size=(X.shape[1], d))

    XW = np.dot(X, W)
    sin_XW, cos_XW = np.sin(XW), np.cos(XW)
    phi = np.concatenate([cos_XW, sin_XW], axis=1)

    return phi



# has to be done at the end, after everything has been imported
import sys

from . import io
from . import tools as tl
from . import utils

sys.modules.update({f"{__name__}.{m}": globals()[m] for m in ["tl"]})



from dataclasses import dataclass, field
from typing import Any, Dict, Iterable, Optional


@dataclass
class JobOutput:
    """
    Output signature of a multiple-processing job. This class provides
    an organized way to record and manipulate the output from a job.


    """

    jobid: int
    run_ok: bool
    result: Optional[Dict[str, Any]] = field(default_factory=dict)

    @property
    def result_keys(self) -> Iterable[str]:
        """keys for the result"""
        return self.result.keys()

    def _as_dict(self, attr_name: str):
        """transform a class attribute into a dict"""
        attr = getattr(self, attr_name)
        if isinstance(attr, dict):
            return attr
        else:
            return {attr_name: attr}



import warnings
from typing import Any, Dict, Literal, Optional, Union

import numpy as np
import scipy
from anndata import AnnData
from scanpy._utils import AnyRandom
from scanpy.tools._utils import _choose_representation
from sklearn.utils import check_random_state

from ._utils import random_feats


def sketch(
    adata: AnnData,
    n_sketch: int,
    use_rep: Optional[str] = None,
    n_pcs: Optional[int] = None,
    random_state: AnyRandom = 0,
    method: Literal["gs", "kernel_herding", "random"] = "gs",
    replace: bool = False,
    inplace: bool = False,
    sketch_kwds: Dict[str, Any] = {},
    key_added: Optional[str] = None,
    copy: bool = False,
) -> Optional[AnnData]:
    """\
    Cell sketching :cite:`Hie19gs` :cite:`Vishal22`

    *sketch* is a function that can be used for summarizing the landscape of
    a large single-cell dataset by selecting a compact subset of cells. This
    function supports three different methods to perform cell sketching:
    geoemetric sketching :cite:`Hie19gs`, kernel herding :cite:`Vishal22`, and simple
    random sampling without replacement.

    .. note::
       More information and bug reports about the geometric sketching method `here
       <https://github.com/brianhie/geosketch>`__.

    Parameters
    ----------
    adata
        Annotated data matrix.
    n_sketch
        number of subsampled cells
    use_rep
        Use the indicated representation. `'X'` or any key for `.obsm` is valid.
        If `None`, the representation is chosen automatically:
        For `.n_vars` < 50, `.X` is used, otherwise 'X_pca' is used.
        If 'X_pca' is not present, it's computed with default parameters.
    n_pcs
        Use this many PCs. If `n_pcs==0` use `.X` if `use_rep is None`
    random_state
        A numpy random seed.
    method
        Use 'gs' :cite:`Hie19gs`, 'kernel_herding' :cite:`Vishal22` or 'random' for
        for summarizing the landscape of large single-cell datasets.
    replace
        whether to sample with replacement (default=False)
    inplace
        It is a boolean which subsample cells in `adata` itself if True.
    sketch_kwds
        A dict that holds other specific parameters used by the sketching methods.
        (e.g. scaling parameter `gamma` and dimensionality of the random Fourier
        frequency features `D` in the kernel herding method.)
    key_added
        If not specified, the subsampled index in boolean format is stored in
        .obs['sketch'], and the sketching parameters are stored in .uns['sketch'].
        If specified, the subsampled index in boolean format is stored in .obs[key_added+'sketch'],
        and the sketching parameters are stored in .uns[key_added+'sketch'].
    copy
        Return a copy instead of writing to adata.

    Returns
    -------
    Depending on `copy`, updates or returns `adata` with the following:

    See `key_added` parameter description for the storage path of subsampled indices.

    Example
    -------
    >>> import scanpy as sc
    >>> import sclkme

    Load annotated dataset:

    >>> adata = sc.datasets.pbmc3k_processed()

    Run the cell sketching using geometric sketching:

    >>> sclkme.tl.sketch(adata, n_sketch=128, use_rep="X_pca", method="gs", key_added = "gs")

    Run the cell sketching using kernel herding:

    >>> sclkme.tl.sketch(adata, n_sketch=128, use_rep="X_pca", method="kernel_herding", key_added = "kh")

    Visualize the sketched dataset:

    >>> adata_sketch_gs = adata[adata.obs['gs_sketch']]
    >>> sc.pl.umap(adata_sketch_gs, color="louvain", size=100)
    >>> adata_sketch_kh = adata[adata.obs['kh_sketch']]
    >>> sc.pl.umap(adata_sketch_kh, color="louvain", size=100)
    """
    # start = logg.info("Cell sketching")

    if method == "gs":
        try:
            from geosketch import gs
        except ImportError:
            raise ImportError(
                "Please install the package of geometric sketching:\n\t `pip install geosketch`."
            )

    adata = adata.copy() if copy else adata
    if adata.is_view:
        adata._init_as_actual(adata.copy())

    if key_added is None:
        key_added = "sketch"
    else:
        key_added = key_added + "_sketch"

    adata.uns[key_added] = {"params": {"n_sketch": n_sketch}, "method": method}
    if use_rep is not None:
        adata.uns[key_added]["params"]["use_rep"] = use_rep
    if n_pcs is not None:
        adata.uns[key_added]["params"]["n_pcs"] = n_pcs

    adata.uns[key_added]["params"]["random_state"] = random_state
    random_state = check_random_state(random_state)

    if n_sketch >= adata.n_obs:
        replace = True
        warnings.warn("n_sketch too large, adjusting to sampling with replacement.")

    # find the representation for sketching
    X = _choose_representation(adata, use_rep=use_rep, n_pcs=n_pcs, silent=True)
    # main routine
    if method == "gs":
        # get the seed from random state
        random_seed = random_state.get_state()[1][0]
        sketch_index = gs(X, n_sketch, seed=random_seed, replace=replace)

    elif method == "kernel_herding":
        if replace:
            raise ValueError(
                "`kernel herding` cannot perform sampling with replacement, "
                "set `rsampling=False` or decrease the size of `n_sketch`."
            )
        if isinstance(X, scipy.sparse.csr_matrix):
            X = X.toarray()

        if sketch_kwds:
            for key in sketch_kwds:
                adata.uns[key_added]["params"][key] = sketch_kwds[key]

        adata.uns[key_added]["params"]["gamma"] = sketch_kwds.get("gamma", 1)
        adata.uns[key_added]["params"]["D"] = sketch_kwds.get("D", 2000)
        phi_X = random_feats(
            X,
            gamma=adata.uns[key_added]["params"]["gamma"],
            D=adata.uns[key_added]["params"]["D"],
            random_state=random_state,
        )
        sketch_index, _ = kernel_herding(X, phi_X, n_sketch)

    elif method == "random":
        size = adata.n_obs
        sketch_index = random_state.choice(
            np.arange(size), size=n_sketch, replace=replace
        )
    else:
        raise ValueError(
            f"method: {method} is not supported, use `gs`, `kernel_herding` or `random` instead."
        )

    # sort the indices
    sketch_index = np.sort(sketch_index)
    if inplace:
        adata = adata[sketch_index, :].copy()
    else:
        adata.obs[key_added] = False
        adata.obs.loc[adata.obs.index[sketch_index], key_added] = True

    # logg.info(
    #     "    finished",
    #     time=start,
    #     deep=f'added column: "{key_added}", sketch mask (bool) in `adata.obs` ',
    # )
    return adata if copy else None


def kernel_herding(X: np.ndarray, phi: np.ndarray, num_samples: int):
    """\
    Computes random fourier frequency features: https://arxiv.org/pdf/2207.00584.pdf

    Parameters
    X: np.ndarray
        array of input data (dimensions = cells x features)
    phi: np.ndarray
        random fourier features generated from the input data: `X`
    num_samples: int
        number of cells to subsample
    ----------
    Returns
    indices: np.ndarray
        indices of sampled cells (dimensions = (`num_samples`, ))
    subsample: np.ndarray
        subsampled `num_samples` cells (dimensions = `num_samples` x features)
    ----------

    """
    w_t = np.mean(phi, axis=0)
    w_0 = w_t
    indices, subsample = [], []
    ind_set = set()
    while len(indices) < num_samples:
        new_ind = np.argmax(np.dot(phi, w_t))
        x_t = X[new_ind]
        w_t = w_t + w_0 - phi[new_ind]
        if new_ind not in ind_set:
            indices.append(new_ind)
            subsample.append(x_t)
            ind_set.add(new_ind)

    indices = np.asarray(indices)
    subsample = np.stack(subsample, axis=0)

    return indices, subsample
